from pathlib import Path

import pytest


@pytest.fixture(name="io_dir", scope="session")
def fixture_io_dir(fixtures_dir: Path) -> Path:
    """Return the path to the io dir."""
    return Path(fixtures_dir, "io")


@pytest.fixture(name="example_json_path", scope="session")
def fixture_example_json_path(fixtures_dir: Path) -> Path:
    """Return a file path to example JSON file."""
    return Path(fixtures_dir, "io", "example_json.json")


@pytest.fixture(name="example_yaml_path", scope="session")
def fixture_example_yaml_path(io_dir: Path) -> Path:
    """Return the path to a example YAML file."""
    return Path(io_dir, "example_yaml.yaml")



"""Module for Tower Open API."""

import logging
import os
from typing import List, Tuple

import requests
from requests import ConnectionError, HTTPError
from requests.exceptions import MissingSchema

from trailblazer.apps.tower.models import (
    TowerProcess,
    TowerTask,
    TowerTaskResponse,
    TowerWorkflowResponse,
)
from trailblazer.constants import TOWER_STATUS, TrailblazerStatus

LOG = logging.getLogger(__name__)


class TowerApiClient:
    """A class handling requests and responses to and from the Tower Open APIs.
    Endpoints are defined in https://tower.nf/openapi/."""

    def __init__(self, workflow_id: str):
        self.workflow_id: str = workflow_id
        self.workspace_id: str = os.environ.get("TOWER_WORKSPACE_ID", None)
        self.tower_access_token: str = os.environ.get("TOWER_ACCESS_TOKEN", None)
        self.tower_api_endpoint: str = os.environ.get("TOWER_API_ENDPOINT", None)
        self.workflow_endpoint: str = f"workflow/{self.workflow_id}"
        self.tasks_endpoint: str = f"{self.workflow_endpoint}/tasks"

    @property
    def headers(self) -> dict:
        """Return headers required for an NF Tower API call.
        Accept and Authorization fields are mandatory."""
        return {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.tower_access_token}",
        }

    @property
    def request_params(self) -> List[Tuple]:
        """Return required parameters for an NF Tower API call.
        Workspace ID is mandatory."""
        return [
            ("workspaceId", self.workspace_id),
        ]

    def build_url(self, endpoint: str) -> str:
        """Build an url to query tower."""
        return self.tower_api_endpoint + endpoint

    def send_request(self, url: str) -> dict:
        """Sends a request to the server and returns the response. NF Tower API calls follow the next schema:
        curl -X GET "<URL>?workspaceId=<WORKSPACE_ID>" \
        -H "Accept: application/json"  \
        -H "Authorization: Bearer <TOWER_ACCESS_TOKEN>
        """
        try:
            response = requests.get(
                url,
                headers=self.headers,
                params=self.request_params,
                verify=True,
            )
            if response.status_code == 404:
                LOG.info("Request failed for url %s\n", url)
                response.raise_for_status()
        except (MissingSchema, HTTPError, ConnectionError) as error:
            LOG.info("Request failed for url %s: Error: %s\n", url, error)
            return {}

        return response.json()

    @property
    def meets_requirements(self) -> bool:
        """Return True if required variables are not empty."""
        if self.tower_api_endpoint is None or self.tower_api_endpoint == "":
            LOG.info("Error: no endpoint specified for Tower Open API request.")
            return False
        if self.tower_access_token is None or self.tower_access_token == "":
            LOG.info("Error: no access token specified for Tower Open API request.")
            return False
        if self.workspace_id is None or self.workspace_id == "":
            LOG.info("Error: no workspace specified for Tower Open API request.")
            return False
        return True

    @property
    def tasks(self) -> TowerTaskResponse:
        """Return a tasks response with information about submitted jobs."""
        if self.meets_requirements:
            url = self.build_url(endpoint=self.tasks_endpoint)
            return TowerTaskResponse(**self.send_request(url=url))

    @property
    def workflow(self) -> TowerWorkflowResponse:
        """Return a workflow response with general information about the analysis."""
        if self.meets_requirements:
            url = self.build_url(endpoint=self.workflow_endpoint)
            return TowerWorkflowResponse(**self.send_request(url=url))


class TowerAPI:
    """Class communicating with NF tower regarding a given analysis (workflow)."""

    def __init__(self, workflow_id: str, dry_run: bool = False):
        self.workflow_id: str = workflow_id
        self.dry_run: bool = dry_run
        self._tower_client = None
        self._response = None
        self._tasks_response = None

    @property
    def tower_client(self) -> TowerApiClient:
        """Returns a NF Tower client."""
        if not self._tower_client:
            self._tower_client = TowerApiClient(workflow_id=self.workflow_id)
        return self._tower_client

    @property
    def response(self) -> TowerWorkflowResponse:
        """Returns a workflow response containing general information about an analysis."""
        if not self._response:
            self._response = self.tower_client.workflow
        return self._response

    @property
    def tasks_response(self) -> TowerTaskResponse:
        """Returns a tasks response containing information about jobs submitted as part of an analysis.
        A task response does not include all jobs that will be run as part of an analysis.
        It only includes jobs that have been submitted a at given time that could be either
        pending, running, completed or failed."""
        if not self._tasks_response:
            self._tasks_response = self.tower_client.tasks
        return self._tasks_response

    @property
    def status(self) -> str:
        """Returns the status of an analysis (also called workflow in NF Tower)."""
        status: str = TOWER_STATUS.get(self.response.workflow.status, TrailblazerStatus.ERROR.value)

        # If the whole workflow (analysis) is completed set it as QC instead of COMPLETE
        if status == TrailblazerStatus.COMPLETED.value:
            return TrailblazerStatus.QC.value
        return status

    @property
    def is_pending(self) -> bool:
        """Returns True if workflow has not started running. Otherwise returns False."""
        return self.status == TrailblazerStatus.PENDING.value

    @property
    def is_complete(self) -> bool:
        """Returns True if workflow has completed. Otherwise returns False."""
        return self.status == TrailblazerStatus.QC.value

    @property
    def processes(self) -> List[TowerProcess]:
        """Returns processes. Processes are steps in a pipeline that might or might not be executed by a given analysis.
        A process can also have more than one corresponding job (task).
        For example, a process could be a 'fastqc' step that could be run multiple times depending on
        how many input files are given."""
        return [TowerProcess(**process) for process in self.response.progress.processesProgress]

    @property
    def tasks(self) -> List[TowerTask]:
        """Returns tasks. Tasks correspond to jobs that have been submitted at a given point regardless of their
        status."""
        return [task["task"] for task in self.tasks_response.tasks]

    @property
    def total_jobs(self) -> int:
        """Returns the total number of jobs for a workflow."""
        return len(self.processes)

    @property
    def succeeded_jobs(self) -> int:
        """Returns the number of succeeded jobs for a workflow."""
        return sum(process.is_complete for process in self.processes)

    @property
    def progress(self) -> float:
        """Returns the progress fraction of a workflow.
        Note this number is not accurate since that exact number of
        processes to be run is unknown."""
        if self.is_complete:
            return 1
        elif self.is_pending or self.total_jobs == 0:
            return 0
        else:
            return round(float(self.succeeded_jobs) / self.total_jobs, 2)

    def get_jobs(self, analysis_id: int) -> List[dict]:
        """Returns a list of jobs associated to a workflow."""
        return [self._get_job(task=task, analysis_id=analysis_id) for task in self.tasks]

    def _get_job(self, task: TowerTask, analysis_id: int) -> dict:
        """Format a job with required information."""
        return dict(
            analysis_id=analysis_id,
            slurm_id=task.nativeId,
            name=task.process,
            status=task.status,
            started_at=task.start,
            elapsed=int(task.duration / 60),
        )



"""Utility functions to simply add test data in a Trailblazer store."""
from trailblazer.store.api import Store
from trailblazer.store.models import Info, User


class StoreHelpers:
    """Class to hold helper functions that needs to be used all over."""

    @staticmethod
    def add_info(store: Store) -> Info:
        """Addd an info object to the store."""
        info: Info = Info()
        store.add(info)
        store.commit()
        return info

    @staticmethod
    def add_user(email: str, name: str, store: Store) -> User:
        """Addd a user object to the store."""
        user: User = User(email=email, name=name)
        store.add(user)
        store.commit()
        return user



from pathlib import Path

import pytest

from trailblazer.constants import FileExtension
from trailblazer.exc import ValidationError
from trailblazer.io.validate_path import validate_file_suffix


def test_validate_file_suffix_correct_suffix(sample_data_path: Path):
    """Test validate file suffix when file suffix is in correct format."""
    # GIVEN a path in the correct file format

    # WHEN validating the file suffix
    was_validated: bool = validate_file_suffix(
        path_to_validate=sample_data_path, target_suffix=FileExtension.YAML
    )

    # THEN assert the suffix is in the correct format
    assert was_validated


def test_validate_file_suffix_wrong_suffix(caplog):
    """Test validate file suffix when file suffix is in wrong format."""
    # GIVEN path in the wrong file format
    file_path: Path = Path("file.json")
    assert file_path.suffix != FileExtension.YAML

    # WHEN validating the file suffix
    with pytest.raises(ValidationError):
        validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtension.YAML)

        # THEN assert the suffix is in the wrong format
        assert "seems to be in wrong format" in caplog.text



import datetime as dt
from pathlib import Path
from typing import Dict, List, Generator

import pytest

from tests.apps.tower.conftest import CaseIDs, TowerTaskResponseFile
from tests.mocks.store_mock import MockStore
from trailblazer.apps.tower.models import TowerTask
from trailblazer.constants import TOWER_TIMESTAMP_FORMAT, TrailblazerStatus, FileFormat
from trailblazer.io.controller import ReadFile
from trailblazer.store.utils.store_helper import StoreHelpers


@pytest.fixture(scope="session", name="username")
def fixture_username() -> str:
    """Return a username."""
    return "Paul Anderson"


@pytest.fixture(scope="session", name="user_email")
def fixture_user_email() -> str:
    """Return an user email."""
    return "paul.anderson@magnolia.com"


@pytest.fixture(scope="session", name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixtures' dir."""
    return Path("tests", "fixtures")


@pytest.fixture(scope="session", name="sample_data_path")
def fixture_sample_data_path(fixtures_dir: Path) -> Path:
    """Return the path to the sample data file."""
    return Path(fixtures_dir, "sample-data.yaml")


@pytest.fixture(name="sample_data")
def fixture_sample_data(sample_data_path: Path) -> Dict[str, list]:
    """Return content of the sample data file."""
    return ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=sample_data_path)


@pytest.fixture(name="trailblazer_tmp_dir")
def fixture_trailblazer_tmp_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for Trailblazer testing."""
    return tmpdir_factory.mktemp("trailblazer_tmp")


@pytest.fixture(name="trailblazer_context")
def fixture_trailblazer_context(sample_store: MockStore) -> Dict[str, MockStore]:
    """Trailblazer context to be used in CLI."""
    return {"trailblazer": sample_store}


@pytest.fixture(name="store")
def fixture_store() -> Generator[MockStore, None, None]:
    """Empty Trailblazer database."""
    _store = MockStore(uri="sqlite://")
    _store.setup()
    yield _store
    _store.drop_all()


@pytest.fixture(name="info_store")
def fixture_info_store(store: MockStore) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated info table."""
    StoreHelpers.add_info(store=store)
    yield store


@pytest.fixture(name="user_store")
def fixture_user_store(
    user_email: str, username: str, store: MockStore
) -> Generator[MockStore, None, None]:
    """A Trailblazer database wih a populated user table."""
    StoreHelpers.add_user(email=user_email, name=username, store=store)
    yield store


@pytest.fixture(name="sample_store")
def fixture_sample_store(
    sample_data: Dict[str, list], store: MockStore
) -> Generator[MockStore, None, None]:
    """A sample Trailblazer database populated with pending analyses."""
    for user_data in sample_data["users"]:
        store.add_user(user_data["name"], user_data["email"])
    for analysis_data in sample_data["analyses"]:
        analysis_data["case_id"] = analysis_data["family"]
        analysis_data["user"] = store.user(analysis_data["user"])
        store.add(store.Analysis(**analysis_data))
    store.commit()
    yield store


@pytest.fixture(name="timestamp_now", scope="session")
def fixture_timestamp_now() -> dt.datetime:
    """Return a time stamp of today's date in date time format."""
    return dt.datetime.now()


@pytest.fixture(name="tower_jobs")
def fixture_tower_jobs(analysis_id, started_at, slurm_id, tower_task_name) -> List[dict]:
    """Return a list of Tower Jobs."""
    return [
        dict(
            analysis_id=analysis_id,
            slurm_id=slurm_id,
            name=tower_task_name,
            started_at=started_at,
            elapsed=63,
            status=TrailblazerStatus.COMPLETED.value,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611829",
            name="NFCORE_RNAFUSION:RNAFUSION:PIZZLY_WORKFLOW:KALLISTO_QUANT",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING.value,
        ),
        dict(
            analysis_id=analysis_id,
            slurm_id="4611828",
            name="NFCORE_RNAFUSION:RNAFUSION:FASTQC",
            started_at=None,
            elapsed=0,
            status=TrailblazerStatus.PENDING.value,
        ),
    ]


@pytest.fixture(name="analysis_id")
def fixture_analysis_id() -> int:
    """Return a mock ID of the analysis in the Trailblazer database."""
    return 1


@pytest.fixture(name="started_at", scope="session")
def fixture_started_at() -> dt.datetime:
    """Returns a started at date."""
    return dt.datetime.strptime("2023-04-04T08:11:27Z", TOWER_TIMESTAMP_FORMAT)


@pytest.fixture(name="slurm_id", scope="session")
def fixture_slurm_id() -> str:
    """Returns a slurm id."""
    return "4611827"


@pytest.fixture(name="tower_task_name", scope="session")
def fixture_tower_task_name() -> str:
    """Returns a NF Tower task name."""
    return "NFCORE_RNAFUSION:RNAFUSION:INPUT_CHECK:SAMPLESHEET_CHECK"


@pytest.fixture(name="case_id", scope="session")
def fixture_case_id() -> str:
    """Return a case ID."""
    return CaseIDs.RUNNING


@pytest.fixture(name="tower_task", scope="session")
def fixture_tower_task() -> TowerTask:
    """Return a Tower task."""
    tower_task_running_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=TowerTaskResponseFile.RUNNING
    )
    return TowerTask(**tower_task_running_content["tasks"][0]["task"])



from datetime import datetime
from typing import List

from trailblazer.constants import TOWER_TIMESTAMP_FORMAT, TOWER_TIMESTAMP_FORMAT_ALTERNATIVE


def datetime_converter(timestamp: str, allowed_formats: List[str]) -> datetime:
    """Converts a timestamp into a datatime object."""
    for dt_format in allowed_formats:
        try:
            return datetime.strptime(timestamp, dt_format)
        except ValueError:
            continue


def tower_datetime_converter(timestamp: str) -> datetime:
    """Converts a NF Tower timestamp into a datatime object."""
    allowed_formats = [TOWER_TIMESTAMP_FORMAT, TOWER_TIMESTAMP_FORMAT_ALTERNATIVE]
    return datetime_converter(timestamp, allowed_formats)



from alchy import Query
from dataclasses import dataclass
from typing import Type

from trailblazer.store.models import Model


@dataclass
class BaseHandler_2:
    """All models in one base class."""

    def __init__(self):
        pass

    def get_query(self, table: Type[Model]) -> Query:
        """Return a query for the given table."""
        return self.query(table)



import subprocess

from tests.apps.tower.conftest import TOWER_ID, CaseIDs, TowerResponseFile, TowerTaskResponseFile
from tests.mocks.tower_mock import MockTowerAPI
from trailblazer.store.api import Store


class MockStore(Store):
    """Instance of Store that mimics workflow manager outputs and interactions."""

    @staticmethod
    def query_slurm(job_id_file: str, case_id: str, ssh: bool) -> bytes:
        """Mock SLURM output."""
        slurm_dict = {
            "blazinginsect": "tests/fixtures/sacct/blazinginsect_sacct",  # running
            "crackpanda": "tests/fixtures/sacct/crackpanda_sacct",  # failed
            "daringpidgeon": "tests/fixtures/sacct/daringpidgeon_sacct",  # failed
            "emptydinosaur": "tests/fixtures/sacct/emptydinosaur_sacct",  # failed
            "escapedgoat": "tests/fixtures/sacct/escapegoat_sacct",  # pending
            "fancymole": "tests/fixtures/sacct/fancymole_sacct",  # completed
            "happycow": "tests/fixtures/sacct/happycow_sacct",  # pending
            "lateraligator": "tests/fixtures/sacct/lateraligator_sacct",  # failed
            "liberatedunicorn": "tests/fixtures/sacct/liberatedunicorn_sacct",  # error
            "nicemice": "tests/fixtures/sacct/nicemice_sacct",  # completed
            "rarekitten": "tests/fixtures/sacct/rarekitten_sacct",  # canceled
            "trueferret": "tests/fixtures/sacct/trueferret_sacct",  # running
        }
        return subprocess.check_output(["cat", slurm_dict.get(case_id)])

    @staticmethod
    def cancel_slurm_job(slurm_id: int, ssh: bool = False) -> None:
        return

    @staticmethod
    def query_tower(config_file: str, case_id: str) -> MockTowerAPI:
        """Return a mocked NF Tower API response."""
        configs = {
            CaseIDs.RUNNING: {
                "workflow_response_file": TowerResponseFile.RUNNING,
                "tasks_response_file": TowerTaskResponseFile.RUNNING,
                "tower_id": TOWER_ID,
                "analysis_id": 1,
            },
            CaseIDs.PENDING: {
                "workflow_response_file": TowerResponseFile.PENDING,
                "tasks_response_file": TowerTaskResponseFile.PENDING,
                "tower_id": TOWER_ID,
                "analysis_id": 1,
            },
            CaseIDs.COMPLETED: {
                "workflow_response_file": TowerResponseFile.COMPLETED,
                "tasks_response_file": TowerTaskResponseFile.COMPLETED,
                "tower_id": TOWER_ID,
                "analysis_id": 1,
            },
        }
        case = configs.get(case_id)
        tower_api = MockTowerAPI(workflow_id=case.get("tower_id"))
        tower_api.mock_query(response_file=case.get("workflow_response_file"))
        tower_api.mock_tasks_query(response_file=case.get("tasks_response_file"))
        return tower_api



from trailblazer.store.base import BaseHandler_2
from trailblazer.store.models import User


class UpdateHandler(BaseHandler_2):
    """Class for updating items in the database."""

    def add_user(self, name: str, email: str) -> User:
        """Add a new user to the database."""
        new_user: User = self.User(email=email, name=name)
        self.add_commit(new_user)
        return new_user



from sqlalchemy.orm import Query

from tests.mocks.store_mock import MockStore
from trailblazer.store.filters.user_filters import filter_users_by_email
from trailblazer.store.models import User


def test_filter_users_by_email(user_store: MockStore, user_email: str):
    """Test return bed by name."""
    # GIVEN a store containing a user

    # WHEN retrieving a user by email
    users: Query = filter_users_by_email(users=user_store.get_query(table=User), email=user_email)

    # ASSERT that the users is a query
    assert isinstance(users, Query)

    # THEN users should be returned
    assert users

    # THEN the users attribute email should match the original
    assert users[0].email == user_email



"""Store backend in Trailblazer."""
import datetime as dt
import io
import logging
import subprocess
from pathlib import Path
from typing import Any, List, Optional

import alchy
import pandas as pd
import sqlalchemy as sqa
from alchy import Query
from dateutil.parser import parse as parse_datestr

from trailblazer.apps.tower.api import TowerAPI
from trailblazer.constants import (
    COMPLETED_STATUS,
    FAILED_STATUS,
    ONGOING_STATUSES,
    SLURM_ACTIVE_CATEGORIES,
    STARTED_STATUSES,
    TrailblazerStatus,
    WorkflowManager,
    FileFormat,
)
from trailblazer.exc import EmptySqueueError, TowerRequirementsError, TrailblazerError
from trailblazer.io.controller import ReadFile
from trailblazer.store.core import CoreHandler
from trailblazer.store.models import Model, User, Analysis, Job
from trailblazer.store.utils import formatters

LOG = logging.getLogger(__name__)


class BaseHandler:
    User = User
    Analysis = Analysis
    Job = Job

    def setup(self):
        self.create_all()

    def get_analysis(self, case_id: str, started_at: dt.datetime, status: str) -> Analysis:
        """
        used in LOG
        Find a single analysis."""
        query = self.Analysis.query.filter_by(family=case_id, started_at=started_at, status=status)
        return query.first()

    def aggregate_failed(self, since_when: dt.date = None) -> List[dict]:
        """
        used in FRONTEND
        Count the number of failed jobs per category (name)."""

        categories = self.session.query(
            self.Job.name.label("name"),
            sqa.func.count(self.Job.id).label("count"),
        ).filter(self.Job.status == "failed")

        if since_when:
            categories = categories.filter(self.Job.started_at > since_when)

        categories = categories.group_by(self.Job.name).all()

        return [{"name": category.name, "count": category.count} for category in categories]

    def analyses(
        self,
        case_id: str = None,
        query: str = None,
        status: str = None,
        deleted: bool = None,
        temp: bool = False,
        before: dt.datetime = None,
        is_visible: bool = None,
        family: str = None,
        data_analysis: str = None,
        comment: str = None,
    ) -> Query:
        """
        used by REST +> CG
        Fetch analyses from the database."""
        if not case_id:
            case_id = family

        analysis_query = self.Analysis.query
        if case_id:
            analysis_query = analysis_query.filter_by(family=case_id)
        if query:  # to be deprecated
            analysis_query = analysis_query.filter(
                sqa.or_(
                    self.Analysis.family.ilike(f"%{query}%"),
                    self.Analysis.status.ilike(f"%{query}%"),
                    self.Analysis.data_analysis.ilike(f"%{query}%"),
                    self.Analysis.comment.ilike(f"%{query}%"),
                )
            )
        if status:
            analysis_query = analysis_query.filter_by(status=status)
        if isinstance(deleted, bool):
            analysis_query = analysis_query.filter_by(is_deleted=deleted)
        if temp:
            analysis_query = analysis_query.filter(self.Analysis.status.in_(ONGOING_STATUSES))
        if before:
            analysis_query = analysis_query.filter(self.Analysis.started_at < before)
        if is_visible is not None:
            analysis_query = analysis_query.filter_by(is_visible=is_visible)
        if data_analysis:
            analysis_query = analysis_query.filter(
                Analysis.data_analysis.ilike(f"%{data_analysis}%")
            )
        if comment:
            analysis_query = analysis_query.filter(Analysis.comment.ilike(f"%{comment}%"))

        return analysis_query.order_by(self.Analysis.started_at.desc())

    def analysis(self, analysis_id: int) -> Optional[Analysis]:
        """
        used by REST
        Get a single analysis by id."""
        return self.Analysis.query.get(analysis_id)

    def get_latest_analysis(self, case_id: str) -> Optional[Analysis]:
        return self.analyses(case_id=case_id).first()

    def get_latest_analysis_status(self, case_id: str) -> Optional[str]:
        """Get the latest analysis status for a case_id"""
        latest_analysis = self.get_latest_analysis(case_id=case_id)
        if latest_analysis:
            return latest_analysis.status

    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        """Check if the latest analysis is ongoing for a case_id"""
        latest_analysis_status = self.get_latest_analysis_status(case_id=case_id)
        return latest_analysis_status in ONGOING_STATUSES

    def is_latest_analysis_failed(self, case_id: str) -> bool:
        """Check if the latest analysis is failed for a case_id"""
        latest_analysis_status = self.get_latest_analysis_status(case_id=case_id)
        return latest_analysis_status == FAILED_STATUS

    def is_latest_analysis_completed(self, case_id: str) -> bool:
        """Check if the latest analysis is completed for a case_id"""
        latest_analysis = self.analyses(case_id=case_id).first()
        return bool(latest_analysis and latest_analysis.status == COMPLETED_STATUS)

    def has_latest_analysis_started(self, case_id: str) -> bool:
        """Check if analysis has started"""
        latest_analysis_status = self.get_latest_analysis_status(case_id=case_id)
        return latest_analysis_status in STARTED_STATUSES

    def add_pending_analysis(
        self,
        case_id: str,
        type: str,
        config_path: str,
        out_dir: str,
        priority: str,
        email: str = None,
        data_analysis: str = None,
        ticket_id: str = None,
        workflow_manager: str = None,
    ) -> Analysis:
        """Add pending entry for an analysis."""
        started_at = dt.datetime.now()
        new_log = self.Analysis(
            family=case_id,
            status="pending",
            started_at=started_at,
            type=type,
            config_path=config_path,
            out_dir=out_dir,
            priority=priority,
            data_analysis=data_analysis,
            ticket_id=ticket_id,
            workflow_manager=workflow_manager,
        )
        new_log.user = self.user(email) if email else None
        self.add_commit(new_log)
        return new_log

    def archive_user(self, user: User, archive: bool = True) -> None:
        """Archive user in the database."""
        user.is_archived = archive
        self.commit()

    def user(self, email: str, include_archived: bool = False) -> User:
        """Fetch a user from the database."""
        query = self.User.query

        if not include_archived:
            query = query.filter_by(is_archived=False)

        query = query.filter_by(email=email)

        return query.first()

    def users(
        self,
        name: str,
        email: str,
        include_archived: bool = False,
    ) -> Query:
        """Fetch all users from the database."""
        query = self.User.query

        if not include_archived:
            query = query.filter_by(is_archived=False)

        if name:
            query = query.filter(self.User.name.contains(name))
        if email:
            query = query.filter(self.User.email.contains(email))

        return query

    def jobs(self) -> Query:
        """Return all jobs in the database."""
        return self.Job.query

    def mark_analyses_deleted(self, case_id: str) -> Query:
        """mark analyses connected to a case as deleted"""
        old_analyses = self.analyses(case_id=case_id)
        if old_analyses.count() > 0:
            for old_analysis in old_analyses:
                old_analysis.is_deleted = True
            self.commit()
        return old_analyses

    def set_analysis_completed(self, analysis_id: int) -> None:
        analysis_obj = self.analysis(analysis_id=analysis_id)
        analysis_obj.status = "completed"
        self.commit()
        LOG.info(f"{analysis_obj.family} - status set to COMPLETED")

    def set_analysis_uploaded(self, case_id: str, uploaded_at: dt.datetime) -> None:
        """Setting analysis uploaded at."""
        analysis_obj: Analysis = self.get_latest_analysis(case_id=case_id)
        analysis_obj.uploaded_at: dt.datetime = uploaded_at
        self.commit()

    def set_analysis_status(self, case_id: str, status: str):
        """Setting analysis status."""
        analysis_obj: Analysis = self.get_latest_analysis(case_id=case_id)
        analysis_obj.status: str = status
        self.commit()
        LOG.info(f"{analysis_obj.family} - Status set to {status}")

    def add_comment(self, case_id: str, comment: str):
        analysis_obj: Analysis = self.get_latest_analysis(case_id=case_id)
        analysis_obj.comment: str = (
            " ".join([analysis_obj.comment, comment]) if analysis_obj.comment else comment
        )

        self.commit()

        LOG.info(f"Adding comment {comment} to analysis {analysis_obj.family}")

    def delete_analysis(self, analysis_id: int, force: bool = False) -> None:
        """Delete the analysis output."""
        analysis_obj = self.analysis(analysis_id=analysis_id)
        if not analysis_obj:
            raise TrailblazerError("Analysis not found")

        if not force and analysis_obj.status in ONGOING_STATUSES:
            raise TrailblazerError(
                f"Analysis for {analysis_obj.family} is currently running! Use --force flag to delete anyway."
            )
        LOG.info(f"Deleting analysis {analysis_id} for case {analysis_obj.family}")
        analysis_obj.delete()
        self.commit()

    @staticmethod
    def query_slurm(job_id_file: str, case_id: str, ssh: bool) -> Any:
        """Args:
        job_id_file: Path to slurm id .YAML file as string
        case_id: Unique internal case identifier which is expected to by the only item in the .YAML dict
        ssh : Whether the request is executed from hasta or clinical-db"""
        job_id: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=Path(job_id_file)
        )
        submitted_job_ids = job_id.get(next(iter(job_id)))
        job_ids_string = ",".join(map(str, submitted_job_ids))
        if ssh:
            return (
                subprocess.check_output(
                    [
                        "ssh",
                        "hiseq.clinical@hasta.scilifelab.se",
                        "squeue",
                        "-j",
                        job_ids_string,
                        "-h",
                        "--states=all",
                        "-o",
                        "%A,%j,%T,%l,%M,%S",
                    ],
                    universal_newlines=True,
                )
                .strip()
                .replace("//n", "/n")
            )
        else:
            return subprocess.check_output(
                [
                    "squeue",
                    "-j",
                    job_ids_string,
                    "-h",
                    "--states=all",
                    "-o",
                    "%A,%j,%T,%l,%M,%S",
                ]
            )

    @staticmethod
    def get_time_elapsed_in_min(elapsed_string: Optional[str]) -> Optional[int]:
        """Parse SLURM elapsed time string into minutes"""
        if not elapsed_string or not isinstance(elapsed_string, str):
            return 0
        days = 0
        if "-" in elapsed_string:
            days = int(elapsed_string.split("-")[0])
            elapsed_string = elapsed_string.split("-")[1]
        split_timestamp = elapsed_string.split(":")
        if len(split_timestamp) < 3:
            split_timestamp = list("0" * (3 - len(split_timestamp))) + split_timestamp
        return int(
            (parse_datestr(":".join(split_timestamp)) - parse_datestr("0:0:0")).seconds / 60
            + days * 24 * 60
        )

    @staticmethod
    def parse_squeue_to_df(squeue_response: Any, ssh: bool) -> pd.DataFrame:
        """Reads queue response into a pandas dataframe for easy parsing.
        Raises:
            TrailblazerError: when no entries were returned by squeue command"""
        if not squeue_response:
            raise EmptySqueueError("No jobs found in SLURM registry")
        if ssh:
            parsed_df = pd.read_csv(
                io.StringIO(squeue_response),
                sep=",",
                header=None,
                na_values=["nan", "N/A", "None"],
                names=["id", "step", "status", "time_limit", "time_elapsed", "started"],
            )
        else:
            parsed_df = pd.read_csv(
                io.BytesIO(squeue_response),
                sep=",",
                header=None,
                na_values=["nan", "N/A", "None"],
                names=["id", "step", "status", "time_limit", "time_elapsed", "started"],
            )
        parsed_df["time_elapsed"] = parsed_df["time_elapsed"].apply(
            lambda x: Store.get_time_elapsed_in_min(x)
        )
        return parsed_df

    def update_slurm_jobs(self, analysis_obj: Analysis, jobs_dataframe: pd.DataFrame) -> None:
        """Parses job dataframe and creates job objects"""
        if len(jobs_dataframe) == 0:
            return
        formatter_func = formatters.formatter_map.get(
            analysis_obj.data_analysis, formatters.transform_undefined
        )
        jobs_dataframe["step"] = jobs_dataframe["step"].apply(lambda x: formatter_func(x))

        for job_obj in analysis_obj.failed_jobs:
            job_obj.delete()
        self.commit()
        analysis_obj.failed_jobs = [
            self.Job(
                analysis_id=analysis_obj.id,
                slurm_id=val.get("id"),
                name=val.get("step"),
                status=val.get("status").lower(),
                started_at=parse_datestr(val.get("started"))
                if isinstance(val.get("started"), str)
                else None,
                elapsed=val.get("time_elapsed"),
            )
            for ind, val in jobs_dataframe.iterrows()
        ]
        self.commit()

    def update_ongoing_analyses(self, ssh: bool = False) -> None:
        """Iterate over all analysis with ongoing status and query SLURM for current progress"""
        ongoing_analyses = self.analyses(temp=True)
        for analysis_obj in ongoing_analyses:
            try:
                self.update_run_status(analysis_id=analysis_obj.id, ssh=ssh)
            except Exception as error:
                LOG.error(
                    f"Failed to update {analysis_obj.family} - {analysis_obj.id}: {type(error).__name__}"
                )

    @staticmethod
    def get_elapsed_time(self, analysis_obj: Analysis) -> str:
        """Get elapsed time for the analysis"""
        return str(
            (
                dt.datetime.now()
                - min(
                    job_obj.started_at for job_obj in analysis_obj.failed_jobs if job_obj.started_at
                )
            )
        )

    def update_run_status(self, analysis_id: int, ssh: bool = False) -> None:
        """Query entries related to given analysis, and update the Trailblazer database."""
        analysis: Analysis = self.analysis(analysis_id)
        if not analysis:
            LOG.warning(f"Analysis {analysis_id} not found!")
            return
        if analysis.workflow_manager == WorkflowManager.TOWER.value:
            self.update_tower_run_status(analysis_id=analysis_id)
        elif analysis.workflow_manager == WorkflowManager.SLURM.value:
            self.update_slurm_run_status(analysis_id=analysis_id, ssh=ssh)

    def update_slurm_run_status(self, analysis_id: int, ssh: bool = False) -> None:
        """Query slurm for entries related to given analysis, and update the Trailblazer database"""
        analysis: Analysis = self.analysis(analysis_id)
        try:
            jobs_dataframe = self.parse_squeue_to_df(
                squeue_response=self.query_slurm(
                    job_id_file=analysis.config_path, case_id=analysis.family, ssh=ssh
                ),
                ssh=ssh,
            )
            self.update_slurm_jobs(analysis_obj=analysis, jobs_dataframe=jobs_dataframe)

            status_distribution = round(
                jobs_dataframe.status.value_counts() / len(jobs_dataframe), 2
            )

            LOG.info(f"Status in SLURM: {analysis.family} - {analysis_id}")
            LOG.info(jobs_dataframe)
            analysis.progress = float(status_distribution.get("COMPLETED", 0.0))
            if status_distribution.get("FAILED") or status_distribution.get("TIMEOUT"):
                if status_distribution.get("RUNNING") or status_distribution.get("PENDING"):
                    analysis.status = TrailblazerStatus.ERROR.value
                else:
                    analysis.status = TrailblazerStatus.FAILED.value

            elif status_distribution.get("COMPLETED") == 1:
                analysis.status = TrailblazerStatus.COMPLETED.value

            elif status_distribution.get("PENDING") == 1:
                analysis.status = TrailblazerStatus.PENDING.value

            elif status_distribution.get("RUNNING"):
                analysis.status = TrailblazerStatus.RUNNING.value

            elif status_distribution.get("CANCELLED") and not (
                status_distribution.get("RUNNING") or status_distribution.get("PENDING")
            ):
                analysis.status = TrailblazerStatus.CANCELLED.value

            LOG.info(f"Updated status {analysis.family} - {analysis.id}: {analysis.status} ")
            self.commit()

            analysis.logged_at = dt.datetime.now()
        except Exception as e:
            LOG.error(f"Error logging case - {analysis.family} : {e.__class__.__name__}")
            analysis.status = "error"
            self.commit()

    @staticmethod
    def query_tower(config_file: str, case_id: str) -> TowerAPI:
        """Parse a config file to extract a NF Tower workflow ID and return a TowerAPI.
        Currently only one tower ID is supported."""
        workflow_id: int = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=Path(config_file)
        ).get(case_id)[0]
        tower_api = TowerAPI(workflow_id=workflow_id)
        if not tower_api.tower_client.meets_requirements:
            raise TowerRequirementsError
        return tower_api

    def update_tower_run_status(self, analysis_id: int) -> None:
        """Query tower for entries related to given analysis, and update the Trailblazer database."""
        analysis: Analysis = self.analysis(analysis_id)
        tower_api: TowerAPI = self.query_tower(
            config_file=analysis.config_path, case_id=analysis.family
        )

        try:
            LOG.info(
                f"Status in Tower: {analysis.family} - {analysis_id} - {tower_api.workflow_id}"
            )
            analysis.status: str = tower_api.status
            analysis.progress: int = tower_api.progress
            analysis.logged_at: dt.datetime = dt.datetime.now()
            self.update_jobs(analysis=analysis, jobs=tower_api.get_jobs(analysis_id=analysis.id))
            self.commit()
            LOG.info(f"Updated status {analysis.family} - {analysis.id}: {analysis.status} ")
        except Exception as error:
            LOG.error(f"Error logging case - {analysis.family} :  {type(error).__name__}")
            analysis.status: str = TrailblazerStatus.ERROR.value
            self.commit()

    def update_jobs(self, analysis: Analysis, jobs: List[dict]) -> None:
        """Updates failed jobs in the analysis."""
        for job in analysis.failed_jobs:
            job.delete()
        analysis.failed_jobs = [self.Job(job) for job in jobs]
        self.commit()

    @staticmethod
    def cancel_slurm_job(slurm_id: int, ssh: bool = False) -> None:
        """Cancel slurm job by slurm job ID"""
        if ssh:
            subprocess.Popen(
                ["ssh", "hiseq.clinical@hasta.scilifelab.se", "scancel", str(slurm_id)]
            )
        else:
            subprocess.Popen(["scancel", str(slurm_id)])

    def cancel_analysis(self, analysis_id: int, email: str = None, ssh: bool = False) -> None:
        """Cancel all ongoing slurm jobs associated with the analysis, and set job status to canceled"""
        analysis_obj = self.analysis(analysis_id=analysis_id)
        if not analysis_obj:
            raise TrailblazerError(f"Analysis {analysis_id} does not exist")

        if analysis_obj.status not in ONGOING_STATUSES:
            raise TrailblazerError(f"Analysis {analysis_id} is not running")

        for job_obj in analysis_obj.failed_jobs:
            if job_obj.status in SLURM_ACTIVE_CATEGORIES:
                LOG.info(f"Cancelling job {job_obj.slurm_id} - {job_obj.name}")
                self.cancel_slurm_job(job_obj.slurm_id, ssh=ssh)
        LOG.info(
            f"Case {analysis_obj.family} - Analysis {analysis_id}: all ongoing jobs cancelled successfully!"
        )
        self.update_run_status(analysis_id=analysis_id)
        analysis_obj.status = "canceled"
        analysis_obj.comment = (
            f"Analysis cancelled manually by user:"
            f" {(self.user(email).name if self.user(email) else (email or 'Unknown'))}!"
        )
        self.commit()


class Store(alchy.Manager, BaseHandler, CoreHandler):
    def __init__(self, uri: str):
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=Model)



from pathlib import Path

from trailblazer.constants import FileFormat
from trailblazer.io.controller import ReadFile


def test_get_content_from_file(example_yaml_path: Path):
    """
    Tests get content from file.
    """
    # GIVEN a YAML file

    # WHEN reading the YAML file
    yaml_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=example_yaml_path
    )

    # Then assert a dict is returned
    assert isinstance(yaml_content, dict)


def test_get_content_from_file_when_json(example_json_path: Path):
    """
    Tests get content from file when JSON.
    """
    # GIVEN a JSON file

    # WHEN reading the JSON file
    json_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=example_json_path
    )

    # Then assert a dict is returned
    assert isinstance(json_content, dict)



from datetime import datetime

from tests.mocks.store_mock import MockStore
from trailblazer.store.models import Info


def test_info_query(info_store: MockStore):
    """Test base model table query."""

    # GIVEN a store

    # WHEN getting an object using the query
    info: Info = info_store.get_query(table=Info).first()

    # THEN it should return an Info object with a default "created_at" date
    assert isinstance(info.created_at, datetime)



from json import JSONDecodeError
from pathlib import Path
from typing import Optional

from trailblazer.apps.tower.api import TowerAPI
from trailblazer.apps.tower.models import TowerTaskResponse, TowerWorkflowResponse
from trailblazer.constants import FileFormat
from trailblazer.io.controller import ReadFile


class MockTowerAPI(TowerAPI):
    """Instance of TowerAPI that mimics expected Tower output."""

    @property
    def response(self) -> Optional[dict]:
        return self.mock_response or None

    @property
    def tasks_response(self) -> Optional[dict]:
        return self.mock_tasks_response or None

    def mock_query(self, response_file: Path) -> TowerWorkflowResponse:
        try:
            self.mock_response = TowerWorkflowResponse(
                **ReadFile.get_content_from_file(
                    file_format=FileFormat.JSON, file_path=response_file
                )
            )
        except JSONDecodeError:
            self.mock_response = TowerWorkflowResponse(**{})
        return self.mock_response

    def mock_tasks_query(self, response_file: Path) -> TowerTaskResponse:
        try:
            self.mock_tasks_response = TowerTaskResponse(
                **ReadFile.get_content_from_file(
                    file_format=FileFormat.JSON, file_path=response_file
                )
            )
        except JSONDecodeError:
            self.mock_tasks_response = TowerTaskResponse(**{})
        return self.mock_tasks_response



from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, validator

from trailblazer.constants import TOWER_PROCESS_STATUS, TOWER_STATUS, TrailblazerStatus
from trailblazer.utils.datetime import tower_datetime_converter

SCALE_TO_MILLISEC: int = 1000


class TowerTask(BaseModel):
    """NF Tower task model."""

    process: str
    name: str
    status: str
    nativeId: str
    dateCreated: Optional[Union[str, datetime]]
    lastUpdated: Optional[Union[str, datetime]]
    start: Optional[Union[str, datetime]]
    duration: Optional[int]
    hash: Optional[str]
    tag: Optional[str]
    submit: Optional[str]
    complete: Optional[str]
    module: List[str]
    container: Optional[str]
    attempt: Optional[int]
    script: Optional[str]
    scratch: Optional[str]
    workdir: Optional[str]
    queue: Optional[str]
    cpus: Optional[int]
    memory: Optional[int]
    disk: Optional[int]
    time: Optional[int]
    env: Optional[str]
    executor: Optional[str]
    machineType: Optional[str]
    cloudZone: Optional[str]
    priceModel: Optional[str]
    cost: Optional[float]
    errorAction: Optional[str]
    exitStatus: Optional[int]
    realtime: Optional[int]
    pcpu: Optional[float]
    pmem: Optional[float]
    rss: Optional[int]
    vmem: Optional[int]
    peakRss: Optional[int]
    peakVmem: Optional[int]
    rchar: Optional[int]
    wchar: Optional[int]
    syscr: Optional[int]
    syscw: Optional[int]
    readBytes: Optional[int]
    writeBytes: Optional[int]
    volCtxt: Optional[int]
    invCtxt: Optional[int]
    exit: Optional[str]
    id: Optional[int]
    taskId: Optional[int]

    class Config:
        validate_all = True

    @validator("duration")
    def set_duration(cls, duration: Optional[int]) -> int:
        """Convert milliseconds to seconds or return 0 if empty."""
        return round(duration / SCALE_TO_MILLISEC) if duration else 0

    @validator("status")
    def set_status(cls, status) -> str:
        return TOWER_STATUS.get(status)

    @validator("start", "dateCreated", "lastUpdated")
    def set_datetime(cls, time) -> Optional[Union[str, datetime]]:
        if type(time) is str:
            return tower_datetime_converter(timestamp=time)
        elif type(time) is datetime:
            return time
        else:
            return None

    @property
    def is_complete(cls) -> bool:
        """Returns if the process succeded."""
        return cls.status == TrailblazerStatus.COMPLETED.value


class TowerProcess(BaseModel):
    """NF Tower task model."""

    process: str
    dateCreated: Optional[Union[str, datetime]]
    lastUpdated: Optional[Union[str, datetime]]
    pending: Optional[int]
    submitted: Optional[int]
    running: Optional[int]
    succeeded: Optional[int]
    failed: Optional[int]
    cached: Optional[int]
    memoryEfficiency: Optional[float]
    cpuEfficiency: Optional[float]
    loadCpus: Optional[int]
    loadMemory: Optional[int]
    peakCpus: Optional[int]
    peakTasks: Optional[int]
    peakMemory: Optional[int]

    class Config:
        validate_all = True

    @validator("dateCreated", "lastUpdated")
    def set_datetime(cls, time) -> Optional[Union[str, datetime]]:
        if type(time) is str:
            return tower_datetime_converter(timestamp=time)
        elif type(time) is datetime:
            return time
        else:
            return None

    @property
    def is_complete(cls) -> bool:
        """Returns if the process succeded."""
        return cls.status == TrailblazerStatus.COMPLETED.value

    @property
    def status(cls) -> str:
        for status_flag in TOWER_PROCESS_STATUS.keys():
            if cls.dict().get(status_flag, 0):
                return TOWER_PROCESS_STATUS.get(status_flag)
        return TrailblazerStatus.ERROR.value


class TowerWorkflow(BaseModel):
    """NF Tower workflow model."""

    status: str


class TowerProgress(BaseModel):
    """NF Tower progress model."""

    workflowProgress: Dict
    processesProgress: List


class TowerTaskResponse(BaseModel):
    """NF Tower task response model."""

    tasks: List[Optional[Dict[str, TowerTask]]]
    total: int


class TowerWorkflowResponse(BaseModel):
    """NF Tower task model."""

    workflow: TowerWorkflow
    progress: TowerProgress



from flask_alchy import Alchy

from trailblazer.store.api import BaseHandler
from trailblazer.store.models import Model


class TrailblazerAlchy(Alchy, BaseHandler):
    pass


store = TrailblazerAlchy(Model=Model)
